unit FFTModule;

interface
uses Transformation;

function FFTInit(X,Y:longword):TTransError; cdecl;
function FFTFinalize : TTransError; cdecl;
function FFTGetBVal(FnX,FnY,X,Y:longword):TFloat; cdecl;
function FFTTransform(Dir:TTransDirection; Data:pointer):TTransError; cdecl;
function FFTGetP(ID:longword;var Value:pointer; var Len:longword):TTransError; cdecl;
function FFTSetP(ID:longword;Value:pointer;Len:longword):TTransError; cdecl;
function FFTGetConf(var Value:pointer):longword; cdecl;
procedure FFTSetConf(Value:pointer;Len:longword); cdecl;
function FFTGetStatus : TTransStatus; cdecl;
procedure FFTGetEM(Mask:pointer; Energy:pointer) cdecl;


var
  FFT : TTransformation = (
    Name : 'FFT';
    Description : 'Fast Discrete Fourier Transformation';
    ID : 1;
    Version : $100;
    Public : false;
    Init : FFTInit;
    Finalize : FFTFinalize;
    GetEnergyMatrix : FFTGetEM;
    GetBasisValue : FFTGetBVal;
    Transform : FFTTransform;
    // property 0: basis layer, type isn't pointer, but 32 bit word (0 for real, 1 for complex layer)
    // property 1: Get> pointer to an array of size X*Y with TFloat elements. Array contains from complex layer of transformation result
    //             Set> sets complex layer for next transformation. Value points to an array X*Y of TFloat elements with complex parts of input data
    GetProperty : FFTGetP;
    SetProperty : FFTSetP;
    GetConfig : FFTGetConf;
    SetConfig : FFTSetConf;
    Configure : nil;
    GetQuantMatrixItem : nil;
    Status : FFTGetStatus;
  );


implementation
uses DOT, Complex;

var
  FFTData : TFFTData;
  CplxLayer : TShapeData;
  PowW1, PowW2 : TFFTData;
  BasisResultLayer : byte;
  AutoZeroCplxLayer : byte;

function FFTInit;
var i:integer;
begin
  //Set layer to real
  BasisResultLayer:=0;
  //Automatically fill complex layer with zeros before forward transformations
  AutoZeroCplxLayer:=1;
  //Initialize internal transformation to FFTII
  FFT.LastError:=TTransError(DOT.Initialize(X,Y,dmFFT));
  if FFT.LastError=teOK then begin
    FFT.X:=X;
    FFT.Y:=Y;
    try
      SetLength(FFTData,X*Y);
      SetLength(CplxLayer,X*Y);
      // create powers of W
      SetLength(PowW1,X);
      SetLength(PowW2,Y);
      for i:=0 to FFT.X*FFT.Y-1 do
        CplxLayer[i]:=0;
      for i:=0 to X-1 do begin
        PowW1[i].r:=cos(2*Pi*i/X);
        PowW1[i].i:=sin(2*Pi*i/X);
      end;
      for i:=0 to Y-1 do begin
        PowW2[i].r:=cos(2*Pi*i/Y);
        PowW2[i].i:=sin(2*Pi*i/Y);
      end;
    except
      DOT.Finalize;
      FFT.LastError:=teNotEnoughMem;
    end;
  end;
  Result:=FFT.LastError;
end;


function FFTFinalize;
begin
  SetLength(FFTData,0);
  SetLength(CplxLayer,0);
  SetLength(PowW1,0);
  SetLength(PowW2,0);
  FFT.LastError:=TTransError(DOT.Finalize);
  Result:=FFT.LastError;
end;


function FFTGetBVal;
var basisVal:TCplxNumber;
begin
  if (DOT.State=dsClosed) then begin FFT.LastError:=teClosed; Result:=0; exit end;
  if (DOT.GetMode<>dmFFT) then begin FFT.LastError:=teIncorrectMode; Result:=0; exit end;
  BasisVal:=CplxMul(PowW1[(FnX*X) mod FFT.X], PowW2[(FnY*Y) mod FFT.Y]);
  basisVal.r:=basisVal.r/(FFT.X*FFT.Y);
  basisVal.i:=basisVal.i/(FFT.X*FFT.Y);
  if BasisResultLayer=0 then Result:=basisVal.r else Result:=basisVal.i;
end;


function FFTTransform;
var i:integer;
begin
  if (DOT.State=dsClosed) then begin FFT.LastError:=teClosed; Result:=teClosed; exit end;
  if (DOT.GetMode<>dmFFT) then begin FFT.LastError:=teIncorrectMode; Result:=teIncorrectMode; exit end;
  case AutoZeroCplxLayer of
    0 : for i:=0 to FFT.X*FFT.Y-1 do begin
          FFTData[i].r:=TShapeData(Data)[i];
          FFTData[i].i:=CplxLayer[i];
        end;
    1 : for i:=0 to FFT.X*FFT.Y-1 do begin
          FFTData[i].r:=TShapeData(Data)[i];
          FFTData[i].i:=0;
        end;
  end;
  FFT.LastError:=TTransError(DOT.Transform(TDOTDirection(Dir),FFTData,nil));
  for i:=0 to FFT.X*FFT.Y-1 do begin
    TShapeData(Data)[i]:=FFTData[i].r;
    CplxLayer[i]:=FFTData[i].i;
  end;
  Result:=FFT.LastError;
end;


function FFTGetP;
begin
  if (DOT.State=dsClosed) then begin FFT.LastError:=teClosed; Result:=teClosed; exit end;
  if (DOT.GetMode<>dmFFT) then begin FFT.LastError:=teIncorrectMode; Result:=teIncorrectMode; exit end;
  case ID of
  0 : begin
    Value:=pointer(BasisResultLayer);
    Len:=0;
    Result:=teOK;
  end;
  1 : begin
    Value:=CplxLayer;
    Len:=FFT.X*FFT.Y*sizeof(CplxLayer[0]);
    Result:=teOK;
  end;
  2 : begin
    Value:=pointer(AutoZeroCplxLayer);
    Len:=0;
    Result:=teOK;
  end;
  else
    Result:=tePropertyNotExists;
  end;
end;


function FFTSetP;
var i:integer;
begin
  if (DOT.State=dsClosed) then begin FFT.LastError:=teClosed; Result:=teClosed; exit end;
  if (DOT.GetMode<>dmFFT) then begin FFT.LastError:=teIncorrectMode; Result:=teIncorrectMode; exit end;
  case ID of
  0 : begin
    BasisResultLayer:=longword(Value);;
    Result:=teOK;
  end;
  1 : begin
    for i:=0 to FFT.X*FFT.Y-1 do
      CplxLayer[i]:=TShapeData(Value)[i];
    Result:=teOK;
  end;
  2 : begin
    AutoZeroCplxLayer:=byte(Value);
    if AutoZeroCplxLayer>1 then AutoZeroCplxLayer:=1;
    Result:=teOK;
  end;
  else
    Result:=tePropertyNotExists;
  end;
end;

function FFTGetConf;
begin
  Value:=nil;
  Result:=0;
end;


procedure FFTSetConf;
begin

end;

function FFTGetStatus : TTransStatus;
begin
  Result:=TTransStatus(DOT.State);
end;


procedure FFTGetEM;
var
    i:integer;
    sum : TFloat;
begin
  if DOT.GetMode<>dmFFT then begin FFT.LastError:=teIncorrectMode; exit; end;

  sum:=0;
  for i:=0 to FFT.Y*FFT.X-1 do sum:=sum + TShapeData(Mask)[i];

  for i:=0 to FFT.X*FFT.Y-1 do begin
    TShapeData(Energy)[i]:=sum;
    CplxLayer[i]:=0;
  end;
end;





end.
