unit ShapeCodec;

interface
uses Transformation;

type
  TShapeData = Transformation.TShapeData;
  TSCDirection = (scForward, scInverse);
  TShapeMask = array of boolean;
  TStatusProc = procedure (Iteration:integer; PSNR:real; Spectrum:TShapeData; Residue:TShapeData;
                           LastAddedX, LastAddedY:word; var Stop:boolean) of object;


//Transforms an image segment with dimension X,Y (X,Y must be powers of 2) stroed in array Data.
//Direction of transformation is in Direction. Forward transformation process terminates, if
//PSNR of coded image will be better or equal to ReqPSNR.
//During forward transformation iterations after every
//InterruptPeriod-th iteration is called StatusProc (if isn't nil).
//StatusProc receives:
//Iteration = number of current iteration
//PSNR = current Peak Signal Noise Ratio in dB
//Spectrum = current spectrum coeficients (NOTE: changes in Spectrum takes effect for computing)
//Residue = current residue of input picture (it is difference between original image and coded image
// (NOTE: changes in Residue takes effect for computing). Array pointer Residue is equal to input array pointer Data.
//Stop = its value is false, changing to true stops iteration process
function Transform(Transf:PTransformation; X,Y:word; Data:TShapeData; Direction:TSCDirection;
                   Mask:TShapeMask;ReqPSNR:real;InterruptPeriod: word; StatusProc:TStatusProc):TTransError;

implementation
uses Math;


function Transform(Transf:PTransformation; X,Y:word; Data:TShapeData; Direction:TSCDirection;
                   Mask:TShapeMask;ReqPSNR:real;InterruptPeriod: word; StatusProc:TStatusProc):TTransError;
var i,j,Iteration,MaxNRG:integer;
    Spectrum, Temp, Energy : TShapeData;
    Stop:boolean; PSNR:real;
    PointNo : word;
    BasisFn1, BasisFn2 : word;
    m,n,z,s : integer;
begin
  //Compute inverse transforamtion (i.m. a simple inverse DCT)
  if Direction=scInverse then begin
    Result:=Transf.Init(X,Y);
    if Result=teOK then begin
      Transf.Transform(tdInverse,Data);
      Result:=Transf.Finalize;
    end;
    exit;
  end;
  try
    SetLength(Energy,X*Y);
    SetLength(Spectrum,X*Y);
    SetLength(Temp,X*Y);
    Result:=Transf.Init(X,Y);

    //Compute energy divisor matrix
    if Result<>teOK then exit;
    PointNo:=0;
    for i:=0 to X*Y-1 do begin
      Temp[i]:=ord(Mask[i]);
      inc(PointNo,ord(Mask[i]));
    end;

    // if transformation module supports energy matrix computation, compute with this function
    if @Transf.GetEnergyMatrix<>nil then
      Transf.GetEnergyMatrix(Temp,Energy)
    else
      //else compute with slow algoritm
      for z:=0 to Y-1 do
        for s:=0 to X-1 do begin
         Energy[z*X+s]:=0;
         for m:=0 to Y-1 do
           for n:=0 to X-1 do
             if Mask[m*X+n] then begin
               PSNR:=Transf.GetBasisValue(s,z,n,m);  //PSNR is now the selected basis value at X,Y
               Energy[z*X+s]:=Energy[z*X+s]+PSNR*PSNR;
           end;
        end;

    //Initialize result spectrum matrix
    for i:=0 to X*Y-1 do Spectrum[i]:=0;

    //Iteration
    Stop:=false; Iteration:=0;
    //Compute initial PSNR
    Data[0]:=Data[0]*ord(Mask[0]);
    PSNR:=Data[0]*Data[0];
    for i:=1 to X*Y-1 do begin
      Data[i]:=Data[i]*ord(Mask[i]);
      PSNR:=PSNR+Data[i]*Data[i];
    end;
    if PSNR<>0 then PSNR:=10*log10((65025.0*PointNo)/PSNR)
    else PSNR:=1e20;
    while not Stop and (PSNR<ReqPSNR) do begin
      //move residuum data to temp area
      Move(Data[0],Temp[0],X*Y*sizeof(Data[0]));
      //transform it to spectrum
      Transf.Transform(tdForward,Temp);
      //find maximal energy difference
      MaxNRG:=0;
      //Set as maximum first basis function with nonzero energy on shape
      while (Energy[MaxNRG]=0) and (maxNRG<X*Y) do inc(MaxNrg);
      for i:=maxNRG+1 to X*Y-1 do
        if (not isNaN(Energy[i])) then
          if (Energy[i]<>0) and (Temp[i]*Temp[i]*Energy[maxNRG]>Temp[maxNRG]*Temp[maxNRG]*Energy[i]) then
            maxNRG:=i;
      //increment energy on selected basis function
      PSNR:=Temp[maxNRG]/(Energy[maxNRG]);
      Spectrum[maxNRG]:=Spectrum[maxNRG]+PSNR;  //PSNR=delta coeficient
      BasisFn1:=maxNRG mod X;
      BasisFn2:=maxNRG div X;
      //substract basis from residue
      for j:=0 to Y-1 do
        for i:=0 to X-1 do
          if Mask[j*X+i] then
            Data[j*X+i]:=Data[j*X+i]-PSNR*Transf.GetBasisValue(BasisFn1,BasisFn2,i,j)
          else Data[j*X+i]:=0;

      //Compute PSNR & ShapeSum
      PSNR:=Data[0]*Data[0];
      for i:=1 to X*Y-1 do PSNR:=PSNR+Data[i]*Data[i];
      if PSNR<>0 then PSNR:=10*log10((65025.0*PointNo)/PSNR)
      else PSNR:=1e20;
      //handle iteration
      inc(Iteration);
      if (InterruptPeriod>0) and (Iteration mod InterruptPeriod = 0) and (@StatusProc<>nil) then
        StatusProc(Iteration,PSNR,Spectrum,Data, BasisFn1, BasisFn2,Stop);
    end;
    Move(Spectrum[0],Data[0],X*Y*sizeof(Data[0]));
    Result:=Transf.Finalize;
    SetLength(Temp,0); Temp:=nil;
    SetLength(Spectrum,0); Spectrum:=nil;
    SetLength(Energy,0); Energy:=nil;
  except
    Transf.Finalize;
    SetLength(Temp,0); Temp:=nil;
    SetLength(Spectrum,0); Spectrum:=nil;
    SetLength(Energy,0); Energy:=nil;
    for i:=0 to X*Y-1 do Data[i]:=0;
    Result:=teNotEnoughMem;
  end;
end;

end.
