unit Shapes;

interface
uses Controls, Classes, Transformation, ShapeCodec, SysUtils, ExtCtrls, Types, Windows, Graphics;
type
    TShapeMask = ShapeCodec.TShapeMask;
    TShapeData = ShapeCodec.TShapeData;
    // 8 bit picture data array
    TPicData = array of byte;
    // map shape ID (color) array
    TMapData = array of longword;
    // 16 bit integer array for quantized spectrum
    TSpecData = array of integer;

    // Shape object
    TShape = class
       //Create shape with color ID ID, Picture point to 8 bit picture data, PicWidth/Mapwidth is number
       //of pixels in line of picture/map, Map points to shape map array, X Y are coordinates of a point in shape
       constructor Create(ID:longword; Picture:TPicData; PicWidht: integer; Map:TMapData; MapWidth:integer; X,Y:integer);
       //Adds point at [X,Y] in picture with ID to shape. if ID of point isn't equal to shape's ID, point is rejected
       function   AddPoint(ID:longword; X,Y:integer):boolean;
       //Loads shape from stream
       procedure  LoadFromStream(Stream:TStream);
       //Saves shape to stream
       procedure  SaveToStream(Stream:TStream);
       //Draws shape boundaries to 2 canvases
       procedure  Draw(PicDrawA,MapDrawA:TCanvas);
       //Returns true, when point at [x,y] is point of shape
       function   MyPixel(x,y:integer):boolean;
       //Encodes shape with Shape Adaptive Coder algorithm. ReqPSNR is required PSNR,
       //if Continue is true and some spectral data are available, encoding continues from current spectrum,
       // else encoding begins from empty spectrum
       // Statusproc is callaed every InterruptPeriod iteration durgin encoding
       procedure  Encode(TransModID, TransID:longword; ReqPSNR:real; Continue:boolean; InterruptPeriod: word; StatusProc:TStatusProc; shapeName:String);
       //Decodes current spectrum and fills picture to Destination
       //Fillmap at [x,y] is true when pixel is filled, false if no
       function   Decode(Destination:TCanvas; FillMap:TShapeMask; shapeName:String):boolean;
      //Draws current spectrum to Canvas with amplification coeficient Intensity
       procedure  DrawSpectrum(Canvas:TCanvas; shapeName:String);
       //Returns true, if is spectrum availvable
       function   HaveSpectrum:boolean;
       //Fills array Data with rounded spectral coeficients by zig-zag scanning.
       //NormType: 0 = None, 1 = Orthogonal, 2 = Orthonormal
       //Result value is number of spectral coeficients
       function   GetSpectrumStream(var Data:TSpecData):word;
       //Sets spectrum from input stream. Spectrum in input stream is scanned by zig-zag scanning and quantized with quality pQQ
       //and transformed with codec pTransID, pTransModID
       procedure  ReadSpectrumStream(s:TStream; pTransModID,pTransID:longword; pQQ, pNT:byte);
     private
       iPicture: TPicData;
       iMap: TMapData;
       iPicW, iMapW : integer;
       iID : longword;
       iSpectrum : TShapeData;
       extCallBack : TStatusProc;
       iPSNR : real;
       iIterations, begIterations : integer;
       iQuantQuality : byte;
       iQuantQualCoef : real;
       iTransModID : longword;
       iTransID : longword;
       iTransf : PTransformation;
       iTransConfLen : integer;
       iTransConf : array of byte;
       iNormCoef : TFloat;
       iNormType : byte;
       function GetWidth:integer;
       function GetHeight:integer;
       function GetRWidth:integer;
       function GetRHeight:integer;
       procedure CallBack(Iteration:integer; PSNR:real; Spectrum:TShapeData; Residue:TShapeData;
                           LastAddedX, LastAddedY:word; var Stop:boolean);
       function GetQuantizationMatrix(X,Y:word):real;
       procedure SetQuantQuality(q:byte);
       procedure SetNormType(T:byte);
       function GetACSpecCoef:integer;
       property QuantMatrix[X,Y:word]:real read GetQuantizationMatrix;
     public
       Left, Right, Top, Bottom : integer;
       PointCount : integer;
     published
       property ID:longword read iID;
       property Width:integer read GetWidth;
       property Height:integer read GetHeight;
       property RealWidth:integer read GetRWidth;
       property RealHeight:integer read GetRHeight;
       property IterationsDone:integer read iIterations;
       property CurrentPSNR:real read iPSNR;
       property QuantQuality : byte read iQuantQuality write SetQuantQuality;
       property ACSpecCoef : integer read GetACSpecCoef;
       property TransModuleID : longword read iTransModID;
       property TransformationID : longword read iTransID;
       property NormType : byte read iNormType write SetNormType;
     end;

implementation
uses EncSelFrm;

constructor TShape.Create;
begin
  inherited Create;
  iID:=ID;
  iPicture:=Picture;
  iMap:=Map;
  iPicW:=PicWidht;
  iMapW:=MapWidth;
  pointCount:=1;
  iSpectrum:=nil;
  Left:=X; Right:=X; Top:=Y; Bottom:=Y;
  iPSNR:=0;
  iIterations:=0;
  QuantQuality:=0;
  iTransModID:=0;
  iTransID:=0;
  iTransConfLen:=0;
  iTransConf:=nil;
  iNormType:=2; iNormCoef:=1;
end;

function TShape.AddPoint;
begin
  if iMap=nil then begin Result:=false; exit; end;
  if iID=ID then begin
    if X<Left then Left:=X;
    if X>Right then Right:=X;
    if Y<Top then Top:=Y;
    if Y>Bottom then Bottom:=Y;
    inc(pointCount);
  end;
  Result:=iID=ID;
end;

procedure TShape.LoadFromStream;
var tmpB:byte;
    tmpI:integer;
begin
  Stream.Read(iID,sizeof(iId));
  Stream.Read(Left,sizeof(Left));
  Stream.Read(Right,sizeof(Right));
  Stream.Read(Top,sizeof(Top));
  Stream.Read(Bottom,sizeof(Bottom));
  Stream.Read(pointCount,sizeof(pointCount));
  Stream.Read(tmpI,sizeof(tmpI)); //emulald a ColorSum valtozo beolvasasat (eloregedett, csak a kompatibilitas miatt)
  Stream.Read(tmpB,sizeof(tmpB)); QuantQuality:=tmpB;
  Stream.Read(iIterations,sizeof(iIterations));
  Stream.Read(iPSNR,sizeof(iPSNR));
  Stream.Read(iTransModID,sizeof(iTransModID));
  Stream.Read(iTransID,sizeof(iTransID));
  Stream.Read(iTransConfLen,sizeof(iTransConfLen));
  Stream.Read(iTransConf,iTransConfLen);
  NormType:=(tmpI shr 1) and 7;
  if ((tmpI and 1) =1) or (tmpI=pointCount) then begin
    if iIterations>0 then begin
      SetLength(iSpectrum,Width*Height);
      Stream.Read(iSpectrum[0],Width*Height*sizeof(iSpectrum[0]));
    end
  end else begin
    SetLength(iSpectrum,Width*Height);
    for tmpI:=0 to Width*Height-1 do
      iSpectrum[tmpI]:=0;
  end;
end;

procedure TShape.SaveToStream;
var tmp : integer;
begin
  Stream.Write(iID,sizeof(iId));
  Stream.Write(Left,sizeof(Left));
  Stream.Write(Right,sizeof(Right));
  Stream.Write(Top,sizeof(Top));
  Stream.Write(Bottom,sizeof(Bottom));
  Stream.Write(pointCount,sizeof(pointCount));
  if HaveSpectrum and (iIterations=0) then
    tmp:=0
  else
    tmp:=1;
  tmp:=tmp or (iNormType shl 1);
  Stream.Write(tmp,sizeof(tmp)); //emulald a ColorSum valtozo kiirasat (eloregedett, csak a kompatibilitas miatt)
                                 //bit 0:
                                 //  1: 0 iteracio = spektrumhiany.
                                 //  0: 0 iteracio = spektrum tele 0-kal
                                 //bit 3-1: Normalizacio tipusa
  Stream.Write(iQuantQuality,sizeof(iQuantQuality));
  Stream.Write(iIterations,sizeof(iIterations));
  Stream.Write(iPSNR,sizeof(iPSNR));
  Stream.Write(iTransModID,sizeof(iTransModID));
  Stream.Write(iTransID,sizeof(iTransID));
  Stream.Write(iTransConfLen,sizeof(iTransConfLen));
  Stream.Write(iTransConf,iTransConfLen);
  if iIterations>0 then
    Stream.Write(iSpectrum[0],Width*Height*sizeof(iSpectrum[0]));
end;


function RoundToPow2(a:word):word;
asm
  xor cl,cl
  mov ax,a
  or  ax,ax
  jnz @loop1
  mov ax,1
  jmp @quit
@loop1:
  inc cl
  shl ax,1
  jnc  @loop1
  dec cl
  dec cl
  mov ax,8000h
@loop2:
  shr ax,1
  dec cl
  jnz @loop2
@quit:
  mov result,ax
end;

function TShape.GetWidth;
begin
  Result:=RoundToPow2(Right-Left);
end;

function TShape.GetHeight;
begin
  Result:=RoundToPow2(Bottom-Top);
end;

function TShape.GetRWidth;
begin
  Result:=Right-Left+1;
end;

function TShape.GetRHeight;
begin
  Result:=Bottom-Top+1;
end;

procedure TShape.Draw;
var x,y:integer;
begin
  PicDrawA.Pen.Color:=clGreen;
  PicDrawA.Brush.Style:=bsClear;
  MapDrawA.Pen.Color:=clGreen;
  MapDrawA.Brush.Style:=bsClear;
  PicDrawA.Rectangle(Left,Top,Right+1,Bottom+1);
  MapDrawA.Rectangle(Left,Top,Right+1,Bottom+1);
  for y:=Top to Bottom do
    for x:=Left to Right do
      if MyPixel(x,y) and not (MyPixel(x-1,y-1) and MyPixel(x-1,y) and MyPixel(x-1,y+1) and
                                MyPixel(x,y-1) and MyPixel(x,y+1) and
                                MyPixel(x+1,y-1) and MyPixel(x+1,y) and MyPixel(x+1,y+1)) then begin
        MapDrawA.Pixels[x,y]:=clBlue;
        PicDrawA.Pixels[x,y]:=clBlue;
      end;
  PicDrawA.Pen.Color:=clRed;
  PicDrawA.Brush.Style:=bsClear;
  MapDrawA.Pen.Color:=clRed;
  MapDrawA.Brush.Style:=bsClear;
  PicDrawA.Rectangle(Left,Top,Left+Width,Top+Height);
  MapDrawA.Rectangle(Left,Top,Left+Width,Top+Height);
end;

function TShape.MyPixel;
begin
  if iMap=nil then begin Result:=false; exit; end;
  Result:=(x>=Left) and (x<=Right) and (y>=Top) and (y<=Bottom) and (iMap[y*iMapW+x]=iID);
end;

procedure TShape.Encode;
var Mask : TShapeMask;
    X,Y,a:integer;
    pomi:integer;
    begSpectrum : TShapeData;
    pomptr:pointer;
    RadoSeg:TBitmap;

begin
    RadoSeg:=TBitmap.Create;
    RadoSeg.Width:=Width;
    RadoSeg.Height:=Height;
    RadoSeg.PixelFormat:=pf24bit;
    for y:=0 to Height-1 do
    for x:=0 to Width-1 do begin
      pomi:=(iPicture[Left+x+(Top+y)*iPicW]) and 255;
      RadoSeg.Canvas.Pixels[x,y]:=(pomi and 255) shl 16 + (pomi and 255) shl 8 + (pomi and 255);
    end;
    RadoSeg.SaveToFile('SEGMENT_ORIGINAL_IMG_'+shapeName+'.BMP');
    RadoSeg.Free();

  if (iTransID<>0) and (iTransID<>TransID) then begin
    SetLength(iSpectrum,0); iSpectrum:=nil; iPSNR:=0;
  end;
  if (iTransID<>TransID) or (iTransModID<>TransModID) then begin
    iTransID:=TransID;
    iTransModID:=TransModID;
  end;
  iTransf:=EncSelFrm.FindTransformation(iTransModID,iTransID);
  if iTransf=nil then exit;
  if (iPicture=nil) or (iMap=nil) then exit;
  if (reqPSNR<=iPSNR) and Continue then exit;
  if iSpectrum=nil then SetLength(iSpectrum,Width*Height);
  if Continue then begin
    begIterations:=iIterations;
    SetLength(begSpectrum,Width*Height);
    Move(iSpectrum[0],begSpectrum[0],Width*Height*sizeof(iSpectrum[0]));
    if @iTransf.SetConfig<>nil then iTransf.SetConfig(iTransConf,iTransConfLen);
    Transform(iTransf,Width,Height,iSpectrum,scInverse,nil,0,0,nil);
  end;
  SetLength(Mask,Width*Height);
  a:=0;
  for y:=0 to Height-1 do
    for x:=0 to Width-1 do begin
      pomi:=(iPicture[Left+x+(Top+y)*iPicW]) and 255;
      Mask[a]:=MyPixel(Left+x,Top+y);
      if Continue then iSpectrum[a]:=pomi-128-iSpectrum[a]
      else iSpectrum[a]:=pomi-128;
      inc(a);
    end;
  extCallBack:=StatusProc;
  if Transform(iTransf,Width,Height,iSpectrum,scForward,Mask,ReqPSNR,InterruptPeriod,CallBack) = teOK then begin
    if @iTransf.GetConfig<>nil then begin
      iTransConfLen:=iTransf.GetConfig(pomptr);
      SetLength(iTransConf,iTransConfLen);
      Move(pomptr^,iTransConf,iTransConfLen);
    end;
    if Continue then begin
      for a:=0 to Height*Width-1 do
        iSpectrum[a]:=iSpectrum[a]+begSpectrum[a];
    end;
  end else
     raise EExternal.Create('External exception in module or not enough memory');
  SetLength(begSpectrum,0); begSpectrum:=nil;
  SetLength(Mask,0); Mask:=nil;
end;

procedure TShape.DrawSpectrum;
var x,y,a:integer;
    maxValue:TFloat;
    Color:integer;
    qMatrixItem : TFloat;
    RadoSpec:TBitmap;
    radoTxtFile : TextFile;

begin
  if iSpectrum<>nil then begin
    a:=0;
    maxValue:=0;
    for y:=0 to Height-1 do
      for x:=0 to Width-1 do begin
        if maxValue<Abs(iSpectrum[a]) then
          maxValue:=Abs(iSpectrum[a]);
        inc(a);
      end;
    a:=0;
    // Try to open the file for writing to
    AssignFile(radoTxtFile, 'SPECTRUM_TXT_'+shapeName+'.TXT');
    ReWrite(radoTxtFile);
    for y:=0 to Height-1 do  begin
      for x:=0 to Width-2 do begin
         Write(radoTxtFile, iSpectrum[a], ',');
         inc(a);
      end;
      Write(radoTxtFile, iSpectrum[a]);
      inc(a);
      WriteLn(radoTxtFile);
    end;
    CloseFile(radoTxtFile);

    a:=0;
    RadoSpec:=TBitmap.Create;
    RadoSpec.Width:=Width;
    RadoSpec.Height:=Height;
    RadoSpec.PixelFormat:=pf24bit;

 //   for y:=0 to Height-1 do
//     for x:=0 to Width-1 do begin
//       color:=Round(ln(abs(iSpectrum[a])+1)/ln(maxValue+1)*255.0);
//       Canvas.Pixels[x,y]:=(color and 255) shl 16 + (color and 255) shl 8 + (color and 255);
//       RadoSpec.Canvas.Pixels[x,y]:=(color and 255) shl 16 + (color and 255) shl 8 + (color and 255);
//        inc(a);
//      end;
  RadoSpec.SaveToFile('SPECTRUM_IMG_'+shapeName+'.BMP');
  RadoSpec.Free();
  end else
    for y:=0 to Height-1 do
      for x:=0 to Width-1 do
        Canvas.Pixels[x,y]:=0;

end;

function TShape.Decode;
var Data:TShapeData;
    x,y, a : integer;
    pomc : byte;
    RadoReconst:TBitmap;
begin
  iTransf:=EncSelFrm.FindTransformation(iTransModID,iTransID);
  if (iSpectrum=nil) or (iTransf=nil) then begin Result:=false; exit; end;
  SetLength(Data,Height*Width);
  Move(iSpectrum[0],Data[0],Width*Height*SizeOf(Data[0]));
  if QuantQuality<>0 then begin
    a:=0;
    for y:=0 to Height-1 do
      for x:=0 to Width-1 do begin
        Data[a]:=Round(iNormCoef*Data[a]/QuantMatrix[x,y])*QuantMatrix[x,y]/iNormCoef;
        inc(a);
      end;
  end;
  if @iTransf.SetConfig<>nil then iTransf.SetConfig(iTransConf,iTransConfLen);
  Result:=Transform(iTransf,Width,Height,Data,scInverse,nil,0,0,nil)=teOK;
  if Result then begin
    a:=0;

    RadoReconst:=TBitmap.Create;
    RadoReconst.Width:=Width;
    RadoReconst.Height:=Height;
    RadoReconst.PixelFormat:=pf24bit;

    for y:=Top to Top+Height-1 do
      for x:=Left to Left+Width-1 do begin
        if ((iMapW*y+x<Length(FillMap)) and (x<iMapW)) and (not FillMap[iMapW*y+x] or MyPixel(x, y)) then begin
          if Round(Data[a])<-128 then pomc:=0
          else if Round(Data[a])>127 then pomc:=255
          else pomc:=Round(Data[a])+128;
          Destination.Pixels[x,y]:=(pomc shl 16) or (pomc shl 8) or pomc;
          RadoReconst.Canvas.Pixels[x-Left,y-Top]:=(pomc shl 16) or (pomc shl 8) or pomc;
          FillMap[iMapW*y+x]:=MyPixel(x,y);
        end;
        inc(a);
      end;
  end;
  SetLength(Data,0);
  RadoReconst.SaveToFile('SEGMENT_DECODED_IMG_'+shapeName+'.BMP');
end;

function TShape.HaveSpectrum;
begin
  Result:=iSpectrum<>nil;
end;

procedure TShape.CallBack;
begin
  iPSNR:=PSNR;
  iIterations:=begIterations+Iteration;
  if @extCallBack<>nil then extCallBack(begIterations+Iteration,PSNR,Spectrum,Residue,LastAddedX,LastAddedY,Stop);
end;

function TShape.GetQuantizationMatrix;
begin
  if iQuantQuality=0 then begin Result:=1; exit end;
  if iTransf=nil then begin
    iTransf:=EncSelFrm.FindTransformation(iTransModID,iTransID);
    if (iTransf=nil) or (@iTransf.GetQuantMatrixItem=nil) then begin Result:=1; exit; end;
  end;
  Result:=iQuantQualCoef*iTransf.GetQuantMatrixItem(X,Y,Width,Height);
end;

procedure TShape.SetQuantQuality;
begin
  if not (q in [0..100]) or (q=iQuantQuality) then exit;
  iQuantQuality:=q;
  if q in [1..50] then
    iQuantQualCoef:=50/q
  else if q in [51..99] then
    iQuantQualCoef:=2-q/50
  else if q=100 then
    iQuantQualCoef:=1;
end;

function TShape.GetSpectrumStream;
var x,y:integer;
    dir:shortint;
    counter,pos:integer;
begin
  if not HaveSpectrum then begin Result:=0; exit; end;
  x:=Width-1;
  y:=Height-1;
  dir:=1;
  counter:=0;

  while (x>=0) and (y>=0) and (Round(iNormCoef*iSpectrum[y*Width+x]/QuantMatrix[x,y])=0) do begin
    inc(counter);
    case dir of
     -1: if x=0 then begin dec(y); dir:=1 end
         else if y=Height-1 then begin dec(x); dir:=1 end
         else begin dec(x); inc(y) end;
     1: if y=0 then begin dec(x); dir:=-1 end
         else if x=Width-1 then begin dec(y); dir:=-1 end
         else begin inc(x); dec(y) end;
    end;
  end;
  counter:=Width*Height-counter;
  SetLength(Data,counter);
  pos:=0; x:=0; y:=0; dir:=-1;
  while pos<counter do begin
    Data[pos]:=Round(iNormCoef*iSpectrum[y*Width+x]/QuantMatrix[x,y]);
    inc(pos);
    case dir of
     -1: if y=Height-1 then begin inc(x); dir:=1 end
         else if x=0 then begin inc(y); dir:=1 end
         else begin dec(x); inc(y) end;
     1: if x=Width-1 then begin inc(y); dir:=-1 end
         else if y=0 then begin inc(x); dir:=-1 end
         else begin inc(x); dec(y) end;
    end;
  end;
  Result:=counter;
end;

function TShape.GetACSpecCoef;
begin
  if HaveSpectrum then Result:=Round(iSpectrum[0]/QuantMatrix[0,0]) else Result:=0;
end;


procedure TShape.ReadSpectrumStream;
var x,y:integer;
    dir:shortint;
    Data : integer;
begin
  iTransModID:=pTransModID;
  iTransID:=pTransID;
  QuantQuality:=pQQ;
  NormType:=pNT;

  SetLength(iSpectrum,Width*Height);
  //uritsd a spektrumot
  for x:=0 to Width*Height-1 do iSpectrum[x]:=0;
  x:=0; y:=0; dir:=-1;
  while s.Position<s.Size do begin
    s.Read(Data,sizeof(Data));
    iSpectrum[y*Width+x]:=(Data*QuantMatrix[x,y])/iNormCoef;
    case dir of
     -1: if y=Height-1 then begin inc(x); dir:=1 end
         else if x=0 then begin inc(y); dir:=1 end
         else begin dec(x); inc(y) end;
     1: if x=Width-1 then begin inc(y); dir:=-1 end
         else if y=0 then begin inc(x); dir:=-1 end
         else begin inc(x); dec(y) end;
    end;
  end;
end;

procedure TShape.SetNormType;
begin
  case T of
    0: begin iNormCoef:=sqrt(Width*Height); iNormType:=0 end;
    1: begin iNormCoef:=1/sqrt(Width*Height); iNormType:=1 end;
    3: begin iNormCoef:=8/sqrt(Width*Height); iNormType:=3 end;
    4: begin iNormCoef:=64/sqrt(Width*Height); iNormType:=4 end;
    else begin iNormCoef:=1; iNormType:=2 end;
  end;
end;

end.
